-- Meaning
--
-- ICFP 2001 programming contest
-- Tom Moertel <tom-icfp2001@moertel.com>

module Meaning where

import Document

type Meaning = [DecoratedChar]
type DecoratedChar = (PropRec, Char)

data PropRec = PropRec { b, em, i, s, tt :: Bool
                       , u, size         :: Int
                       , color           :: Color
                       }
               deriving (Eq)

instance Show PropRec where
    showsPrec _ pr = 
        ((concatMap (\ (acc,t,f) -> if acc pr then t else f)
          [ (b, "B", ""), (em, "E", ""), (i, "I", "")
          , (s, "S", ""), (tt, "T", "") ]
          ++ (let uval = u pr in if uval > 0
              then "u" ++ show (u pr) else "")
          ++ (let sval = size pr in if sval /= 4
              then "s" ++ show (size pr) else "")
          ++ "c" ++ show (fromEnum $ color pr)) ++)


rootProps :: PropRec
rootProps = PropRec { u = 0
                    , b = False, em = False, i = False, s = False, tt = False
                      -- the two following values are special
                      -- since no document can ever use these particular values
                      -- we can test them to see if a document has changed
                      -- the size or color
                    , size = -1
                    , color = AnyColor
                    }


meaningDiff :: Document -> Document -> [String]
meaningDiff  docA docB =
    zipWith mdiff (docMeaning docA) (docMeaning docB)
        where mdiff :: DecoratedChar -> DecoratedChar -> String
              mdiff a b = (if a /= b then show a ++ "*" else "") ++ show b

docMeaning :: Document -> Meaning
docMeaning doc = ws $ dm rootProps doc []


dm :: PropRec -> Document -> Meaning -> Meaning
dm r (Document dpts) = dm' r dpts

dm' :: PropRec -> [DocPart] -> Meaning -> Meaning
dm' _ [] acc             = acc
dm' r ((DText s):xs) acc = map ((,) r) s ++ dm' r xs acc
dm' r (x:xs) acc         = 
    dm r' doc' $ dm' r xs acc
    where (r', doc') = case x of
                       DB doc  -> (r {b=True}, doc)
                       DEM doc -> (r {em= if s r then em r else not (em r)}
                                  , doc)
                       DI doc  -> (r {i=True}, doc) 
                       DPL doc -> (r {u=0,s=False,em=False,i=False,
                                      b=False,tt=False}, doc)
                       DS doc  -> (r {s=True,em=False}, doc)
                       DTT doc -> (r {tt=True}, doc)
                       DU doc  -> (r {u = let uval = u r in
                                      if uval < 3 then uval + 1 else uval}
                                  , doc)
                       DSize sz doc -> (r {size=sz}, doc)
                       DColor col doc -> (r {color=col}, doc)
                       DText _ -> error 
                                  "DText shouldn't ever match in dm' r (x:xs)"

-- adjust a Meaning based on whitespace context rules
--
-- Validation table
--
--             Ysp-TT     Ysp-notTT    notYsp-TT  notYsp-notTT
--          --------------------------------------------------         
-- Xtt      |  ok         ok           ok         ok
-- notXtt   |                          ok         ok

ws :: Meaning -> Meaning
ws []             = []
ws [x]            = if isWS x then [spaceContext x] else [x]
ws (x:y:zs)       = if not (isWS x)
                    then x : ws (y:zs)
                    -- x is whitespace
                    else let x' = spaceContext x in
                         if isTT x' then x' : ws (y:zs) -- don't trim WS in TT
                         else if not (isWS y) then x' : y : ws zs
                         -- y is also whitespace
                         else let y' = spaceContext y in
                              if x' == y'
                              then ws (x':zs) -- identical -> trim y
                              else x' : y' : ws zs -- otherwise keep y

isWS :: DecoratedChar -> Bool
isWS = ((`elem` " \t\n\r") . snd)

isTT :: DecoratedChar -> Bool
isTT = tt . fst

-- converts a decorated whitespace char into its space-context equivalent

spaceContext :: DecoratedChar -> DecoratedChar
spaceContext (r, c) =
    (r { s = False, em = False, i = False, b = False
       , color = if u r == 0 then White else color r }
     , if tt r then c else ' ')


-- determines whether two PropRecs are equivalent
-- note:  when s is set, em is irrelevant

propEq :: PropRec -> PropRec -> Bool
propEq x y =
    (   b x     == b y
     && i x     == i y
     && s x     == s y
     && (em x   == em y         || s x)
     && tt x    == tt y
     && u x     == u y
     && size x  == size y
     && color x == color y
    )
