-- Parse
--
-- ICFP 2001 programming contest
-- Tom Moertel <tom-icfp2001@moertel.com>

module Parse (parseS) where

import Parsec
import List
import FiniteMap

import Document


docParser :: Parser Document
docParser =
    do 
    dparts <- many $
              do { do { text <- many1 (noneOf "<")
                      ; return (DText text)
                      }
                   <|>
                   do { try (char '<' >> notFollowedBy (char '/'))
                      ; choice tagParsers
                      }
                 }
    return (Document dparts)

tagConstructors :: [(String, Document -> DocPart)]
tagConstructors
     = [ ("B"  , DB)
       , ("EM" , DEM)
       , ("I"  , DI)
       , ("PL" , DPL)
       , ("S"  , DS)
       , ("TT" , DTT)
       , ("U"  , DU)
       , ("0"  , DSize 0)
       , ("1"  , DSize 1)
       , ("2"  , DSize 2)
       , ("3"  , DSize 3)
       , ("4"  , DSize 4)
       , ("5"  , DSize 5)
       , ("6"  , DSize 6)
       , ("7"  , DSize 7)
       , ("8"  , DSize 8)
       , ("9"  , DSize 9)
       , ("r"  , DColor Red)
       , ("g"  , DColor Green)
       , ("b"  , DColor Blue)
       , ("c"  , DColor Cyan)
       , ("m"  , DColor Magenta)
       , ("y"  , DColor Yellow)
       , ("k"  , DColor Black)
       , ("w"  , DColor White)
       ]

mkTagParser :: (String, Document -> DocPart) -> Parser DocPart
mkTagParser (tag, ctor) =
    do string tag
       char '>'
       doc <- docParser
       string "</"
       string tag
       char '>'
       return (ctor doc)
    
tagParsers :: [Parser DocPart]
tagParsers = map (mkTagParser) tagConstructors

parseS :: String -> Either ParseError Document
-- parseS = parse docParser "given string"
parseS = Right . parseX


--
-- HAND ROLLED PARSER
--



parseX :: String -> Document
parseX = Document . fst . xparse

type SymbolMap = FiniteMap String  ([DocPart] -> DocPart)

tagMap :: SymbolMap
tagMap = listToFM [ (str, tc . Document) | (str, tc) <- tagConstructors ]

-- no need to check for correctness of input documents; the
-- specification requires that they be correct

xparse :: String -> ([DocPart], String)
xparse []           = ([], "")
xparse ('<':'/':xs) = ([], tail (dropWhile (/= '>') xs))
xparse ('<':xs) =
    case xs of
    a:'>':xs'   -> xp' [a] xs'
    a:b:'>':xs' -> xp' [a,b] xs'
    xs'         -> parseError $ takeWhile (/= '>') xs'
    where xp' tag xs' = 
              case (lookupFM tagMap tag) of
              Nothing   -> parseError tag
              Just ctor -> let (chlds,xs'')  = xparse xs'
                               (sibs, xs''') = xparse xs''
                               in ((ctor $! chlds) : sibs, xs''')
xparse xs = case break (=='<') xs of { (txt,xs') ->  -- always okay
            let (sibs,xs'') = xparse xs' in (DText txt : sibs, xs'') }
                                                      
parseError tag = error ("parse error: unknown tag <" ++ tag ++ ">")
