-- PropRecUtils
--
-- ICFP 2001 programming contest
-- Tom Moertel <tom-icfp2001@moertel.com>

module PropRecUtils where

import Document
import Meaning

-- PROPERTY RECORD MANIPULATORS

set_b, set_em, set_i, set_s, set_tt :: Bool  -> PropRec -> PropRec
set_u, set_size                     :: Int   -> PropRec -> PropRec
set_color                           :: Color -> PropRec -> PropRec

set_b     x r   = r {b=x}
set_em    x r   = r {em=x}
set_i     x r   = r {i=x}
set_s     x r   = r {s=x}
set_tt    x r   = r {tt=x}
set_u     x r   = r {u=x}
set_size  x r   = r {size=x}
set_color x r   = r {color=x}

-- I have this notion of building "up" a level on property w.r.t.  a
-- target instance of that property.  For booleans, this means going
-- from False to True, regardless of the target instance.  For
-- underline it means adding an underline, again, regardless of the
-- target (provided we won't exceed the maximum).  For Size and Color,
-- however, we go "up to" the target's size or color.

up_b, up_em, up_i, up_s, up_tt      :: Bool  -> PropRec -> PropRec
up_u, up_size                       :: Int   -> PropRec -> PropRec
up_color                            :: Color -> PropRec -> PropRec

up_b     _ r   = r {b=True}
up_em    _ r   = r {em=True} -- or would toggling be better?
up_i     _ r   = r {i=True}
up_s     _ r   = r {s=True}
up_tt    _ r   = r {tt=True}
up_u     _ r   = let ur = u r
                     in if ur < 3 then r {u=(ur + 1)} else error "underline>3"
up_size  x r   = r {size=x}
up_color x r   = r {color=x}

data PropRecEntry =
    forall a . (Eq a) => PRE  (PropRec -> a)                   -- getter
                              (a -> PropRec -> PropRec)        -- setter
                              (a -> PropRec -> PropRec)        -- up-setter
                              (PropRec -> PropRec ->
                               Maybe (Document -> DocPart))    -- up-builder

-- the up-builder below creates a document element that makes a
-- change that effects in a Document the transition that the
-- up-setter makes in a PropRec

propTable :: [ PropRecEntry ]          -- Frm To -> returns ctor when you can
                                       --           build up from Frm to To
propTable =  [ PRE b    set_b    up_b    (\_ r' -> onlyWhen (b r') DB)
             , PRE em   set_em   up_em   (\_ r' -> onlyWhen (em r') DEM)
             , PRE i    set_i    up_i    (\_ r' -> onlyWhen (i r') DI)
             , PRE s    set_s    up_s    (\_ r' -> onlyWhen (s r') DS)
             , PRE tt   set_tt   up_tt   (\_ r' -> onlyWhen (tt r') DTT)
             , PRE u    set_u    up_u    (\r r' -> onlyWhen (u r < u r') DU)
             , PRE size set_size up_size (\_ r' -> onlyWhen (size r' >= 0)
                                                       (DSize (size r')))
             , PRE color set_color up_color
                                         (\_ r' ->
                                          onlyWhen (color r' /= AnyColor)
                                                       (DColor (color r')))
             ]
    where onlyWhen test val = if test then Just val else Nothing

propCompare :: PropRecEntry -> PropRec -> PropRec -> Bool
propCompare pre r =
    case pre of PRE getter _ _ _ ->
                    let val = getter r in \r' -> val == getter r'

propUpBuilder :: PropRecEntry -> PropRec->PropRec-> Maybe (Document -> DocPart)
propUpBuilder pre r r' = 
    case pre of PRE _ _ _ upbuilder -> upbuilder r r'

propCopy :: PropRecEntry -> PropRec -> PropRec -> PropRec
propCopy pre r r' =
    case pre of PRE get set _ _ -> set (get r) r'

propUp :: PropRecEntry -> PropRec -> PropRec -> PropRec
propUp pre r r' =
    case pre of PRE get _ up _ -> up (get r) r'


-- Does the PropRec r sufficiently match the DecoratedChar m so that
-- no markup is required to effect a transition from the decoration
-- described by the r to that of the m?

matchQ :: PropRec -> DecoratedChar -> Bool
matchQ r m@(mr,_) =
    r == mr || closeEnoughForWS r m

-- White space is allowed to differ from the surrounding text provided
-- that the the only differences between it and the context decoration
-- are in s, em, i, and color (and color only if its u is 0).  This is
-- because all white space occurs in a special ``space context'' that
-- sets s, em, i, and (if u==0) u automatically; therefore, there is
-- no need for us to force these settings with markup.

closeEnoughForWS :: PropRec -> DecoratedChar -> Bool
closeEnoughForWS r m =
    let mr = fst m in
    isWS m && and [ -- must be equal in all respects except
                    -- s, em, i, and (iff u==0) color
                    tt mr    == tt r
                  , size mr  == size r
                  , u mr     == u r
                  , color mr == color r || u mr == 0
                  ]
